#include "sr.c"
#include "crc.c"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#include <time.h>


// return codes aliases
#define SUCCESS 0
#define ERROR   1

#define WINDOW_SIZE 16
#define DATA_SIZE 505
#define ACKBUFLEN 4

int main(int argc, char *argv[]) {
  // file descriptor
  int file;
  
  // buffer to create the payload of the UDP packet
  char* buffer;
  char* ack_buffer;
   
  // one-char buffer
  char tmp[1] = "";

  // counter for data length
  int i = DATA_SIZE;
  
  // counter for window size (short int because 2 bytes)
  int seq_num = 0;
  int l = 0;

  // window pointers
  struct st_window_element *we_current;
  window_element window_first = window_element_empty;
  window_element window_last = window_element_empty;
  window_element window_tmp, window_tmp2;
  
  // command-line parameters
  char* FileNameToSend;
  char* ServerIPAddressOrDNSName;
  int ServerUDPPortNumber;
  
  // for the size of the file
  struct stat my_stat;
  
  // datagram socket
  int sockfd;
  struct sockaddr_in my_sa; // connector's address information
  struct hostent *server;
  int numbytes;
  int addr_len;
  
  // to indicate if EOF reached
  int finished_sending_packet = 0;
  
  // misc
  int window_size_actual = 0;
  int seq_num_ack0, seq_num_ack1, seq_num_ack;
  int found;
  char crc = 0;
  int time_last_received = time(NULL);
  
  // parse the command line
  if (argc != 4) {
    printf("Usage: %s [FileNameToSend] [ServerIPAddressOrDNSName] [ServerUDPPortNumber]\n", argv[0]);
    return ERROR;
  }
  
  // assigning parameters
  FileNameToSend = argv[1];
  ServerIPAddressOrDNSName = argv[2];
  ServerUDPPortNumber = atoi(argv[3]);
  
  // valid server?
  if ((server = gethostbyname(argv[2])) == NULL) {  // get the host info
    printf("Invalid server \"%s\"\n", argv[2]);
    return ERROR;
  }
  
  // instantiate the socket
  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    printf("Unable to open the socket.\n");
    return ERROR;
  }
  
  // to set to non-blocking
  fcntl(sockfd, F_SETFL, O_NONBLOCK);
  
  // setting the parameters for the connection
  my_sa.sin_family = AF_INET;     // host byte order
  my_sa.sin_port = htons(ServerUDPPortNumber); // short, network byte order
  my_sa.sin_addr = *((struct in_addr *)server->h_addr);
  memset(&(my_sa.sin_zero), '\0', 8); // zero the rest of the struct

  addr_len = sizeof(struct sockaddr);
  
  // allocating memory for the buffers
  buffer = (char*) malloc(DATA_SIZE+7);
  strcpy(buffer, "");
  ack_buffer = (char*) malloc(4);
  strcpy(ack_buffer, "");
  
  //open the file  
  if ((file = open(FileNameToSend, O_RDONLY)) >= 0) {
    
    // let's init the CRC table
    init_crc_table();
  
    // displaying the file size
    fstat(file, &my_stat);
    printf("The file has a size of %d bytes.\n", (int) my_stat.st_size);
    
    // MAIN LOOP
    do  {
      // do we have to send a packet?
      if (window_size_actual < WINDOW_SIZE && finished_sending_packet == 0) {
        // copying the Sequence Number into buffer
        // first 8 bits
        buffer[0]= seq_num >> 8;
        // last 8 bits
        buffer[1]= seq_num & 0xff;
        
        // copying the data into buffer
        i = 0; // reinitializing i
        while(i < DATA_SIZE && finished_sending_packet == 0 ) {
          if (read(file, tmp, 1)) {
            buffer[5+i]= tmp[0];
            i++;
          }
          else {
            // reached EOF
            // updating EOF (End Of File) Flag and Padding Size 
           buffer[5+i]= 0;
           finished_sending_packet = 1;
          }
        }
        
        // filling EOF (End Of File) Flag and Padding Size
        if (finished_sending_packet == 1) { // we have reached EOF
          // DEBUG
          puts("EOF reached => next packet is the last one.");
          
          buffer[2] = 1;        // EOF
          buffer[3] = i >> 8;   // Padding Size first 8 bits
          buffer[4] = i & 0xff; // Padding Size last 8 bits
        }
        else {
          buffer[2]= 0;
          buffer[3]= (DATA_SIZE - 1) >> 8;   // Padding Size first 8 bits
          buffer[4]= (DATA_SIZE - 1) & 0xff; // Padding Size last 8 bits
        }
      
        // fill the rest of the data with NULL character (0) if necessary
        while(i < DATA_SIZE) {
          buffer[5+i]= 0;
          i++;
        }
        
        // computing CRC
        
        // let's put 0 in the CRC fields
        // 1st 8 bits
        buffer[510]= 0;
        // last 8 bits
        buffer[511]= 0;
        
        // let's compute the CRC of the whole packet with 0 in CRC fields
        crc = crc_calc(0, buffer, 512);
        
        // DEBUG
        // printf("CRC16: %c\n", crc);
        
        // copyinig CRC into packet
        // 1st 8 bits
        buffer[510]= crc >> 8;
        // last 8 bits
        buffer[511]= crc & 0xff;
        
        // filling the struct
        we_current = (st_window_element *) malloc(sizeof(struct st_window_element));
        we_current->seq_num = seq_num;
        we_current->acked = 0;
        
        // copying data into packet
        for(l = 0; l < DATA_SIZE + 7; l++) {
          we_current->pkt[l] = buffer[l];
        }
        
        // let's update window_last
        window_last = window_first;
        while (window_last != window_element_empty && window_last->next != window_element_empty) window_last = window_last->next;
        
        // let's add the element to the window
        if (window_first == window_element_empty) {
          // first element
          window_first = we_current;
          window_last = we_current;
          window_first -> next = window_element_empty;
        }
        else {
          window_last->next = we_current;
          window_last = window_last->next;
          window_last->next = window_element_empty;
        }
        
        window_size_actual++;
        
        // put time_sent to current time
        we_current->time_sent = time(NULL);
        
        // let's send the packet
        if ((numbytes = sendto(sockfd, we_current->pkt, sizeof(we_current->pkt), 0, (struct sockaddr *)&my_sa, sizeof(struct sockaddr))) == -1) {
          perror("sendto");
          return ERROR;
        }
        
        // DISPLAY
        printf("SENT sequence number %d\n", seq_num);
        
        // updating seq num
        // modulo 2^16=65536
        seq_num = (seq_num + 1) % 65536;
      }
      
      // let's wait for ACK
      if ((numbytes=recvfrom(sockfd, ack_buffer, ACKBUFLEN, 0, (struct sockaddr *)&my_sa, &addr_len)) >= 0) {
        ack_buffer[numbytes] = '\0';
        time_last_received = time(NULL);
        
        // extract sequence number
        if (ack_buffer[0] < 0) {
          seq_num_ack0 = 256 + (int) ack_buffer[0];
        }
        else {
          seq_num_ack0 = (int) ack_buffer[0];
        }
        if (ack_buffer[1] < 0) {
          seq_num_ack1 = 256 + (int) ack_buffer[1];
        }
        else {
          seq_num_ack1 = (int) ack_buffer[1];
        }
        seq_num_ack = (seq_num_ack0 * 256) + seq_num_ack1;

        printf("RECEIVED ACK for sequence number %d\n", seq_num_ack);
        
        // reinitialize found
        found = 0;
        
        // look for the element in the window with this seq num
        window_tmp = window_first;
        window_tmp2 = window_element_empty;
        while (window_tmp != window_element_empty && found == 0) {
          if (window_tmp->seq_num == seq_num_ack) {
            found = 1;
          }
          else {
            window_tmp2 = window_tmp;
            window_tmp = window_tmp->next;
          }
        }
      if (found == 1) {
        window_tmp->acked = 1;
        
        // we have to erase the first elements of the window acked (maybe 0)
        window_tmp = window_first;
        
        while(window_tmp != window_element_empty && window_tmp->acked == 1) {
          // let's erase the element
          window_first = window_tmp -> next;
          free(window_tmp);
          window_tmp = window_tmp->next;
          window_size_actual--;
        }
      }
    } //if (found == 1)
      // look for timeouts
      window_tmp = window_first;
      while(window_tmp != window_element_empty) {
      	 if (time(NULL) > (window_tmp->time_sent + 1) && window_tmp->acked == 0) {
      	   // TIMEOUT!
      	   // let's resend the packet
           if ((numbytes = sendto(sockfd, window_tmp->pkt, sizeof(window_tmp->pkt), 0, (struct sockaddr *)&my_sa, sizeof(struct sockaddr))) == -1) {
             perror("sendto");
             return ERROR;
           }
           
           // retrieve sequence number
           if (window_tmp->pkt[0] < 0) {
             seq_num_ack0 = 256 + (int) window_tmp->pkt[0];
           }
           else {
             seq_num_ack0 = (int) window_tmp->pkt[0];
           }
           if (window_tmp->pkt[1] < 0) {
             seq_num_ack1 = 256 + (int) window_tmp->pkt[1];
           }
           else {
             seq_num_ack1 = (int) window_tmp->pkt[1];
           }
           seq_num_ack = (seq_num_ack0 * 256) + seq_num_ack1;
           
           // update time_sent to current time
           window_tmp->time_sent = time(NULL);
           
           // display
           printf("TIMEOUT on sequence number %d (-> resent)\n", seq_num_ack);
      	 }
      	 window_tmp = window_tmp->next;
      }
    } while ((time_last_received + 60 > time(NULL)) && ( (finished_sending_packet == 0) || (window_first != window_element_empty) ));
    
    // let's close the socket
    close(sockfd);
    
  }
  else {
    printf("The file \"%s\" doesn't exist.\n", FileNameToSend);
    return ERROR;
  }
  
  if (time_last_received + 60 == time(NULL)) {
    puts("Transfer aborted: nothing received for 60 seconds.");
    return ERROR;
  }
  else {
    puts("File successfully sent.");
  }
  
  // let's close the file
  close(file);
  
  return SUCCESS;
}
