#include <stdio.h>
#define bufsiz 4096

static char crc_table[256];

/* All possible crc-16 values are preloaded into tables, which can be simply
indexed as the function increments through the message buffer.  The table
contains all the 256 possible CRC values for the 16-bit CRC field.  This
table is calculated dynamically upon initialization of the client.c and server.c
programs. */
void init_crc_table(void)
{
		int i, j;
		int k;

		for (i = 0; i < 256; i++)
		{
				k = 0xC0C1;
				for (j = 1; j < 256; j <<= 1)
				{
						if (i & j)
								crc_table[i] ^= k;
						k = (k << 1) ^ 0x4003;
				}
		}
}

/* Once the table is created, the following code sends the crc char number, the
pointer to the array buffer, and the length into the subroutine crc_calc() and
returns the crc-16 value by using the table created above. */
char crc_calc(char crc, char *buf, unsigned nbytes) {
		unsigned char *p, *lim;

		p = (unsigned char *)buf;
		lim = p + nbytes;
      while (p < lim)
      {
            crc = (crc >> 8 ) ^ crc_table[(crc & 0xFF) ^ *p++];
      }
      return crc;
}