#include "sr.c"
#include "crc.c"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>

// return codes aliases
#define SUCCESS 0
#define ERROR   1

#define PKTBUFLEN 513
#define ACKBUFLEN 4
#define DATA_SIZE 505

int main(int argc, char *argv[]) {
  // file descriptor
  int file;
  
  // buffer to receive the packet
  char* buffer;
  
  // buffer to create the ack
  char* ack;
  
  // one-char buffer
  char tmp[1] = "";
  
  // command-line parameters
  char* FileNameToReceive;
  int UDPPortNumber;
  
  // datagram socket
  int sockfd;
  struct sockaddr_in my_addr; // my address information
  struct sockaddr_in my_sa;   // connector's address information
  int numbytes;
  int addr_len;
  
  // counter
  int l;
  
  // to compute the padding size
  int ps0, ps1, ps;
  
  // to check the CRC16
  int crc0, crc1, crc, crc_bis;
  
  // to retrieve the sequence number
  int seq_num0, seq_num1, seq_num;
  
  // window pointers
  struct st_window_element *we_current;
  struct st_window_element *we_current_bis;
  window_element window_first = window_element_empty;
  window_element window_tmp, window_tmp2;
  
  // misc
  int seq_num_tmp;
  int window_size = 0;
  int last_seq_num_written = -1;
  int time_last_received = time(NULL);
  
  // counter
  int i;
  
  // if packet received can be added to the window
  int ok;
  
  // to flag the last packet
  int last_packet_not_received = 0;
  
  // parse the command line
  if (argc != 3) {
    printf("Usage: %s [FileNameToReceive] [UDPPortNumber]\n", argv[0]);
    return ERROR;
  }
  
  // allocating memory for the buffers
  buffer = (char*) malloc(PKTBUFLEN);
  strcpy(buffer, "");
  ack = (char*) malloc(ACKBUFLEN);
  strcpy(ack, "");
  
  FileNameToReceive = argv[1];
  UDPPortNumber = atoi(argv[2]);
        
  if ((file = open(FileNameToReceive, O_CREAT|O_TRUNC|O_RDWR, S_IREAD | S_IWRITE)) >= 0) {

    if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
        perror("socket");
        exit(1);
    }
    
    // to set to non-blocking
    fcntl(sockfd, F_SETFL, O_NONBLOCK);

    my_addr.sin_family = AF_INET;         // host byte order
    my_addr.sin_port = htons(UDPPortNumber);     // short, network byte order
    my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
    memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct

    if (bind(sockfd, (struct sockaddr *)&my_addr,
		sizeof(struct sockaddr)) == -1) {
        perror("bind");
        exit(1);
    }
    
    addr_len = sizeof(struct sockaddr);
    
    // let's init the CRC table
    init_crc_table();
    
    // add the first record not acked
    window_first = (st_window_element *) malloc(sizeof(struct st_window_element));
    window_first->seq_num = 0;
    window_first->acked = 0;
    window_first->next = window_element_empty;
    window_size++;
    
  while ((time_last_received + 60 > time(NULL)) && ( (last_packet_not_received == 0) || (window_first != window_element_empty) )) {
    
    if ((numbytes=recvfrom(sockfd, buffer, PKTBUFLEN , 0, (struct sockaddr *)&my_sa, &addr_len)) >= 0) {
    buffer[numbytes] = '\0';
    time_last_received = time(NULL);

    // extract sequence number
    if (buffer[0] < 0) {
      seq_num0 = 256 + (int) buffer[0];
    }
    else {
      seq_num0 = (int) buffer[0];
    }
    if (buffer[1] < 0) {
      seq_num1 = 256 + (int) buffer[1];
    }
    else {
      seq_num1 = (int) buffer[1];
    }
    seq_num = (seq_num0 * 256) + seq_num1;

    if (seq_num <= last_seq_num_written) {
      // this packet has been well received before, so we assume
      // that the ACK has been lost and we have to resend the ACK
      
      // let's send ack
      ack[0] = seq_num >> 8;
      ack[1] = seq_num  & 0xff;
            
      // let's compute the CRC16
      ack[2] = 0;
      ack[3] = 0;
      
      // let's compute the CRC of the whole packet with 0 in CRC fields
      crc = crc_calc(0, ack, 512);
      
      // DEBUG
      // printf("CRC16: %c\n", crc);
    
      // copyinig CRC into packet
      // 1st 8 bits
      ack[2]= crc >> 8;
      // last 8 bits
      ack[3]= crc & 0xff;
      
      // let's send the ACK!
      if ((numbytes = sendto(sockfd, ack, 4, 0, (struct sockaddr *)&my_sa, sizeof(struct sockaddr))) == -1) {
        perror("sendto");
        return ERROR;
      }
    }
    else {
    printf("RECEIVED seq_num %d\n", seq_num);
    
    // retrieving padding size
    if (buffer[3] < 0) {
      ps0 = 256 + (int) buffer[3];
    }
    else {
      ps0 = (int) buffer[3];
    }
    if (buffer[4] < 0) {
      ps1 = 256 + (int) buffer[4];
    }
    else {
      ps1 = (int) buffer[4];
    }
    ps = (ps0 * 256) + ps1;    
    
    // retrieving CRC
    if (buffer[510] < 0) {
      crc0 = 256 + (int) buffer[510];
    }
    else {
      crc0 = (int) buffer[510];
    }
    if (buffer[511] < 0) {
      crc1 = 256 + (int) buffer[511];
    }
    else {
      crc1 = (int) buffer[511];
    }
    crc = (crc0 * 256) + crc1;
    
    buffer[510] = 0;
    buffer[511] = 0;
    
    // let's compute the CRC of the whole packet with 0 in CRC fields
    crc_bis = crc_calc(0, buffer, 512);
    if (crc_bis < 0) crc_bis = 65536 + crc_bis;
    
    // DEBUG
    //printf("CRC16: received:%c - computed: %c\n", crc, crc_bis);
    //printf("CRC16: received:%d - computed: %d\n", crc, crc_bis);
    
    // if CRC valid, let's send ACK; otherwise, discard it    
    if (crc == crc_bis) {
      // let's send ack
      ack[0] = buffer[0];
      ack[1] = buffer[1];
      
      // let's compute the CRC16
      ack[2] = 0;
      ack[3] = 0;
      
      // let's compute the CRC of the whole packet with 0 in CRC fields
      crc = crc_calc(0, ack, 512);
      
      // DEBUG
      // printf("CRC16: %c\n", crc);
    
      // copyinig CRC into packet
      // 1st 8 bits
      ack[2]= crc >> 8;
      // last 8 bits
      ack[3]= crc & 0xff;
      
      // let's add the packet to the window
      
      // let's create the element
      we_current = (st_window_element *) malloc(sizeof(struct st_window_element));
      we_current->seq_num = seq_num;
      we_current->acked = 1;
      for(i=0; i < 512; i++) {
        we_current->pkt[i]= buffer[i];
      }
      
      ok = 0;
      
      // let's find the right place
      window_tmp = window_first;
      window_tmp2 = window_element_empty;
      while(window_tmp != window_element_empty && window_tmp->seq_num != seq_num) {
        window_tmp2 = window_tmp;
        window_tmp = window_tmp->next;
      }
      
      if (window_tmp != window_element_empty && window_tmp->seq_num == seq_num) { // we have found its place
        // let's change the element
        we_current->next = window_tmp->next;
        if (window_tmp2 != window_element_empty) {
          window_tmp2->next = we_current;
        } else {
          window_first = we_current;
          window_tmp = window_first;
        }
        window_size++;
        ok = 1;
      }
      else { // window empty
      	if (window_tmp2 != window_element_empty) {
          seq_num_tmp = window_tmp2->seq_num + 1;
        }
        else {
          seq_num_tmp = last_seq_num_written + 1;
        }
        while (seq_num_tmp < seq_num && window_size < (16 - 1)) {
          // add a new record not acked
          we_current_bis = (st_window_element *) malloc(sizeof(struct st_window_element));
          we_current_bis->seq_num = seq_num_tmp;
          we_current_bis->acked = 0;
          we_current_bis->next = window_element_empty;
          if (window_tmp2 == window_element_empty) {
            window_first = we_current_bis;
            window_tmp = window_first->next;
            window_tmp2 = window_first;
          }
          else {
            window_tmp2->next = we_current_bis;
            window_tmp2 = window_tmp2->next;
            window_tmp = window_tmp2->next;
          }
          seq_num_tmp = (seq_num_tmp + 1) % 65536;
          window_size++;
        }
        if (seq_num_tmp == seq_num) { // we can insert it
          we_current->next = window_tmp;
          if (window_tmp2 == window_element_empty) {
            window_first = we_current;
          }
          else {
            window_tmp2->next = we_current;
          }
          ok = 1;
          window_size++;
        }
      }
      
      if (ok == 0) {     
      	//let's free the memory and discard the packet
        free(we_current);
      }
      else {
        // let's send the ACK!
        if ((numbytes = sendto(sockfd, ack, 4, 0, (struct sockaddr *)&my_sa, sizeof(struct sockaddr))) == -1) {
          perror("sendto");
          return ERROR;
        }
        
        // retrieve sequence number
        if (ack[0] < 0) {
          seq_num0 = 256 + (int) ack[0];
        }
        else {
          seq_num0 = (int) ack[0];
        }
        if (ack[1] < 0) {
          seq_num1 = 256 + (int) ack[1];
        }
        else {
          seq_num1 = (int) ack[1];
        }
        seq_num = (seq_num0 * 256) + seq_num1;
        
        // display
        printf("ACK sent for sequence number %d\n", seq_num);
        
        // is it the last packet?
        if (we_current->pkt[2] == 1) {
          last_packet_not_received = 1;
        }
        
        // maybe we can copy packets data in the file
        window_tmp = window_first;
        window_tmp2 = window_element_empty;
        while ((window_tmp != window_element_empty) && (window_tmp->acked == 1)) {
          // let's append it to the file
          // is it last packet?
          if (window_tmp->pkt[2] == 0) {
            for(l = 5; l < DATA_SIZE + 5 ; l++) {
              tmp[0] = window_tmp->pkt[l];
              write(file, tmp, 1);
            }
          }
          else { //last packet
            for(l = 5; l < (ps + 5); l++) {
              tmp[0] = window_tmp->pkt[l];
              write(file, tmp, 1);
            }
          }
          
          last_seq_num_written = window_tmp->seq_num;
          
          // let's remove the element
          if (window_tmp2 == window_element_empty) {
            window_first = window_tmp->next;
            // let's free the memory
            free(window_tmp);
            window_tmp = window_first;
          } else {
            window_tmp2->next = window_tmp->next;
            // let's free the memory
            free(window_tmp);
            window_tmp = window_tmp2->next;
          }
          window_size--;
        }
      }
    }
    else {
      printf(" - packet corrupted!!!!\n");
    }
    }
  }
}

  close(sockfd);

}
else {
  puts("Unable to create file.");
  return ERROR;
}
  
  if (time_last_received + 60 == time(NULL)) {
    puts("Transfer aborted: nothing received for 60 seconds.");
    return ERROR;
  }
  else {
    puts("File successfully received.");
  }

  // let's close the file
  close(file);

  return SUCCESS;
}
